#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import wx

from Table       import Table
from Relation    import Relation
from FileOps     import FileOps
from globalStuff import *

tables=[]           # hier landen die Table()-Objekte
relations=[]        # hier landen die Relation()-Objekte
filename=""         # Dateiname der aktuell geladenen Datei



# ----------------------------------------------------------------------
# Eine kleine Klasse, um neue Dateien via Drag&Drop öffnen zu können.
# In der aufrufenden Klasse muss die Funktion fileDropped() existieren.
class FileDrop(wx.FileDropTarget):
    def __init__(self, window):
        wx.FileDropTarget.__init__(self)
        self.window=window
        self.Ctrl_pressed=False

    def OnDropFiles(self, x, y, filenames):
        self.window.fileDropped(filenames)

    def OnDragOver(self, x, y, d):
        if d==wx.DragCopy:
            if wx.GetKeyState(wx.WXK_CONTROL):
                self.Ctrl_pressed=True
                return(wx.DragLink)
            self.Ctrl_pressed=False
            return(wx.DragCopy)
        return(d)

# ----------------------------------------------------------------------
#
class ERMFrame(wx.Frame):
    def __init__(self, parent, title):
        super(ERMFrame, self).__init__(parent, title=title, size=(1350, 750))
        self.InitUI()

    # ------------------------------------------------------------------
    # Fenter initialisieren
    def InitUI(self):
        self.leftDown=None
        self.left_down_table=None
        self.left_down_delta=(0, 0)
        self.left_down_cornermark=None
        self.errors_in_file=None
        self.mouse_posXY_left_click=None
        self.mouse_posXY_current=None
        self.scroll_distXY=None

        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)

        self.statusbar=self.CreateStatusBar(5)
        self.statusbar.SetStatusWidths([0, 100, 100, -1, -1])

        self.__createMenu()
        
        self.filename_dt=FileDrop(self)     # Drag&Drop von Track-Dateien aktivieren
        self.SetDropTarget(self.filename_dt)

        self.Bind(wx.EVT_CONTEXT_MENU,  self.onContextMenu)
        self.Bind(wx.EVT_PAINT,         self.onPaint)
        self.Bind(wx.EVT_MOTION,        self.onMotion)
        self.Bind(wx.EVT_LEFT_UP,       self.onLeftUp)
        self.Bind(wx.EVT_LEFT_DOWN,     self.onLeftDown)
        self.Bind(wx.EVT_LEFT_DCLICK,   self.onLeftDclick)
        self.Bind(wx.EVT_MOUSEWHEEL,    self.onMouseWheel)
        self.Centre()
        wx.CallLater(10, self.SetFocus)
        wx.CallLater(100, self.loadFile)

    # ------------------------------------------------------------------
    # Stellt ein Raster auf dem DC an.
    def showRaster(self):
        #self.dc.SetPen(wx.Pen("#cccccc", width=1, style=wx.LONG_DASH))
        self.dc.SetPen(wx.Pen("#eeeeee", width=1, style=wx.SOLID))
        sx, sy=self.GetSize()
        for x in range(int(sx/RASTER)+1):
            self.dc.DrawLine(x*RASTER, 0, x*RASTER, sy)
        for y in range(int(sy/RASTER)+1):
            self.dc.DrawLine(0, y*RASTER, sx, y*RASTER)
        self.dc.SetPen(wx.BLACK_PEN)

    # ------------------------------------------------------------------
    # Lädt die Datei gemäß der globalen Variable filename.
    def loadFile(self):
        if filename=="" or not os.path.isfile(filename):
            return
        file_ops=FileOps(tables, relations)
        file_ops.load(filename)
        self.errors_in_file=len(file_ops.log)>0
        if self.errors_in_file:
            lns=""
            for ln in file_ops.log:
                lns+=ln+"\n"
            wx.MessageBox(lns, "Warnings / Errors")
            file_ops.log=[]
        self.statusbar.SetStatusText("tables: %d"%(len(tables),), 1)
        self.statusbar.SetStatusText("relations: %d"%(len(relations),), 2)
        if len(tables)==0 and len(relations)==0:
            self.SetTitle("ERMzilla  (empty or bad file)")
        else:
            self.SetTitle("ERMzilla  (%s)"%(filename,))
        self.Refresh()

    # ------------------------------------------------------------------
    # Löscht die internen Datenstrukturen.
    def unloadFile(self):
        global relations, tables
        for r in relations:
            del r
        for t in tables:
            del t
        relations=[]
        tables=[]

    # ------------------------------------------------------------------
    # Wird aus FileDrop aufgerufen, wenn ein oder mehrere DateiObjekte
    # auf dem Fenster "fallengelassen" wurden.
    def fileDropped(self, filenames):
        global filename
        if len(filenames)!=1:
            wx.MessageBox("dropping multiple files is not allowed", "Error")
            return
        self.unloadFile()
        filename=filenames[0]
        self.loadFile()

    # ------------------------------------------------------------------
    # Darstellung der Tabellen und Relationen.
    def onPaint(self, event):
        self.dc=wx.AutoBufferedPaintDC(self)
        self.dc.SetBackground(wx.Brush("#FFFFFF"))
        self.dc.Clear()
        self.dc.SetTextForeground("#000000")
        if self.raster_mode:
            self.showRaster()

        for t in tables:
            t.calcSize(self.dc)
            t.drawTable(self.dc)
        for r in relations:
            fl=r.getFullList()
            self.dc.DrawLines(fl)
            if r.border_start=="L":
                sx, _=self.dc.GetTextExtent(r.rel_type[0])
                self.dc.DrawText(r.rel_type[0], fl[0][0]-5-sx, fl[0][1]-10)
            else:
                self.dc.DrawText(r.rel_type[0], fl[0][0]+5, fl[0][1]-10)
            if r.border_end=="L":
                sx, _=self.dc.GetTextExtent(r.rel_type[1])
                self.dc.DrawText(r.rel_type[1], fl[-1][0]-5-sx, fl[-1][1]-10)
            else:
                self.dc.DrawText(r.rel_type[1], fl[-1][0]+5, fl[-1][1]-10)
            
            if self.rel_points_mode:
                for p in fl:
                    self.dc.DrawCircle(p[0], p[1], 2)

    # ------------------------------------------------------------------
    # Linke Maustaste wurde geklickt.
    # Klick auf Relations-Eckpunkt bewegt den Eckpunkt.
    # Klick auf Relations-Linie kopiert deren Namen in die StatusBar.
    # Klick auf Tabelle bewegt die Tabelle.
    def onLeftDown(self, event):
        self.leftDown=event.GetPosition()
        for r in relations:                                             # zuerst mal alle Eckpunkte testen
            cm=r.isPointOnCornerMark(self.leftDown)
            if cm:
                self.left_down_cornermark=(r, cm)
                return
        for r in relations:
            cm=r.isPointOnThisRelation(self.leftDown)
            if cm:                                                      # es wurde auf eine Relation geklickt
                if cm[0]=="LN":                                         # und zwar auf eine Linie
                    self.statusbar.SetStatusText("relation: %s"%(r.name,), 3)
                    return
                if cm[0]=="CM":                                         # und zwar auf einen Eckpunkt
                    self.left_down_cornermark=(r, cm[1])
                    return
        for t in tables:
            if t.isUnderXY(self.leftDown):                              # es wurde auf eine Tabelle geklickt
                self.left_down_table=t
                self.left_down_delta=(self.leftDown[0]-t.pos[0], self.leftDown[1]-t.pos[1])
                return
        self.mouse_posXY_left_click=self.leftDown                       # wenn weder auf einer Tabelle, noch einer Relation geklickt wurde...
        self.mouse_posXY_current=self.leftDown                          # ...dann soll wohl der Schirm verschoben werden

    # ------------------------------------------------------------------
    # Linke Maustaste wurde losgelassen.
    def onLeftUp(self, event):
        self.leftDown=None
        self.left_down_table=None
        self.left_down_delta=(0, 0)
        self.left_down_cornermark=None
        self.mouse_posXY_left_click=None
        for t in tables:                                                # Tabellen am Raster ausrichten
            t.pos=pointOnRaster(t.pos)
        for r in relations:                                             # und die Relationen ebenso
            for i, p in enumerate(r.line):
                r.line[i]=pointOnRaster(p)
        self.Refresh()

    # ------------------------------------------------------------------
    # Der MausCursor wird über den DC bewegt.
    def onMotion(self, event):
        cur_pos=event.GetPosition()
        #self.statusbar.SetStatusText("pos: %s"%(cur_pos,), 4)
        if self.leftDown:
            x, y=cur_pos
            if self.left_down_table:                                    # Tabelle wird bewegt
                x1=x-self.left_down_delta[0]
                y1=y-self.left_down_delta[1]
                self.left_down_table.moveAbs((x1, y1))
            elif self.left_down_cornermark:                             # Eckpunkt wird bewegt
                self.left_down_cornermark[0].MoveCornerMarkOnIndex((x, y), self.left_down_cornermark[1])
            elif self.mouse_posXY_left_click:                           # DC wird bewegt
                self.scroll_distXY=cur_pos-self.mouse_posXY_current
                self.mouse_posXY_current=cur_pos
                for t in tables:
                    t.moveRel(self.scroll_distXY)
                for r in relations:
                    for i, p in enumerate(r.line):
                        r.line[i]=(p[0]+self.scroll_distXY[0], p[1]+self.scroll_distXY[1])
            self.Refresh()

    # ------------------------------------------------------------------
    # Linke Maustaste wurde doppel-geklickt.
    # Doppelklick auf Relations-Linie fügt einen neuen Eckpunkt ein
    # Doppelklick auf Relations-Eckpunkt löscht den Eckpunkt
    def onLeftDclick(self, event):
        x, y=event.GetPosition()
        for r in relations:
            fl=r.getFullList()
            for p in fl:
                self.dc.DrawCircle(p[0], p[1], 3)
            a=r.isPointOnThisRelation((x, y))
            if a:
                if a[0]=="LN":
                    r.insertCornerMarkAfterIndex((x, y), a[1])
                    self.Refresh()
                    break
                elif a[0]=="CM":
                    r.removeCornerMarkOnIndex(a[1])
                    self.Refresh()
                    break

    # ------------------------------------------------------------------
    # Wenn sich unter dem MausCursor ein TableConnector befindet, führt
    # das Drehen am Scrollrad zum Wechsel der Tabellen-Seite.
    def onMouseWheel(self, event):
        point=event.GetPosition()
        for r in relations:
            tc=r.isPointOnTableConnector(point)
            if tc:
                if tc[0]==0:    # start
                    r.border_start=r.getOtherBorder(r.border_start)
                else:
                    r.border_end=r.getOtherBorder(r.border_end)
                self.Refresh()
                return

    # ------------------------------------------------------------------
    # Stellt das PopUp-Menü dar.
    def onContextMenu(self, event):
        if os.path.isfile(filename) and not self.errors_in_file:
            self.menue.Enable(100, True)
        else:
            self.menue.Enable(100, False)
        self.PopupMenu(self.menue)
        self.Refresh()

    # ------------------------------------------------------------------
    # Erstellt das PopUp-Menü.
    def __createMenu(self):
        self.raster_mode=True
        self.rel_points_mode=True
        self.menue=wx.Menu()
        self.menue.Append(90, 'reload')
        self.menue.Append(100, 'save')
        self.menue.Append(500, 'raster on/off', "", True)
        self.menue.Append(510, 'relation points on/off', "", True)
        self.Bind(wx.EVT_MENU, self.men_reload, id=90)
        self.Bind(wx.EVT_MENU, self.men_save, id=100)
        self.Bind(wx.EVT_MENU, self.men_raster, id=500)
        self.menue.Check(500, self.raster_mode)
        self.Bind(wx.EVT_MENU, self.men_rel_points, id=510)
        self.menue.Check(510, self.rel_points_mode)
        if not os.path.isfile(filename):
            self.menue.Enable(100, False)

    # ------------------------------------------------------------------
    # Menü - reload
    def men_reload(self, event):
        self.unloadFile()
        self.loadFile()
        self.Refresh()

    # ------------------------------------------------------------------
    # Menü - save
    def men_save(self, event):
        strg_d=""
        strg_p=""
        for t in tables:
            sd, sp=t.write()
            strg_d+=sd
            strg_p+=sp
            for r in relations:
                if r.start_col.split(".")[0]==t.name:
                    sd, sp=r.write()
                    strg_d+=sd
                    strg_p+=sp
        with open(filename, "wt") as fl:
            if strg_d.startswith("\n"):
                fl.write(strg_d[1:])
            else:
                fl.write(strg_d)
            fl.write("\n\n\n")
            fl.write(strg_p)

    # ------------------------------------------------------------------
    # Menü - Raster an/aus
    def men_raster(self, event):
        self.raster_mode=not self.raster_mode

    # ------------------------------------------------------------------
    # Menü - Darstellung der Eckpunkte an/aus
    def men_rel_points(self, event):
        self.rel_points_mode=not self.rel_points_mode


# ----------------------------------------------------------------------
# 
if __name__=='__main__':
    if len(sys.argv)>1:
        ff=os.path.isfile(sys.argv[1])
        if ff:
            filename=sys.argv[1]

    app=wx.App(False)
    #frame=ERMFrame(None, pos=(-1, -1), size=(-1, -1))
    if filename=="":
        title="ERMzilla  (no file loaded)"
    else:
        title="ERMzilla  (%s)"%(filename,)
    frame=ERMFrame(None, title=title)
    frame.Show(True)
    app.MainLoop()
